# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains functions for derivative matrix computation.

 - compute_aniso_dmats and compute_iso_dmats takes care of obtaining 
   flux dependent or flux independent derivative coefficients,
 - prepare_mag_data is an auxiliary function to calculate magnetic flux gradients and
 - px_norm is a simple auxiliary function to calculate norms for coefficients
"""
import numpy as np
from scipy import sparse
from scipy.sparse import spdiags, eye


def compute_iso_dmats(grid, derivatives=[1, 2], mask=None):
    """
    Parameters
    ----------
    grid : RegularGrid
    mask : np.ndarray, optional
    derivatives : list of int/str, optional

    Returns
    -------
    list of dmat pairs
    """
    if mask is None:
        mask = np.ones(grid.size, dtype=np.bool)
    if isinstance(derivatives, int):
        derivatives = [derivatives]
    dmats = []
    for d in derivatives:
        dmats.append(generate_iso_dmat(grid, derivative=d))
    dmats_red = reduce_dmats(dmats, mask.flatten())
    # dmats_dict = dict()
    # for t in tvec:
    #     dmats_dict[t] = dmats_red
    return dmats_red

def compute_aniso_dmats(grid, magflux, derivative=4, mask=None):
    # TODO change smoothing type to string
    """
    Prepares matrix of finite differences, magnetic configuration dependent or
    fixed simple differences. First or second derivative can be used

    Parameters
    ----------
    grid : RegularGrid
        reconstrution grid parameters
    magflux : numpy.ndarray
        values of psi normalized mapped to grid
    mask : numpy.ndarray, optional
        bool mask 
    derivative : int in {1,2,3,4}
        specifies type of derivative scheme, default is 4

    Returns
    -------
    tuple of scipy.sparse.csrmatrix pairs
        derivative matrix
    """
    # tidx = np.searchsorted(self.magfield['time'], self.tvec)
    # for t in self.tvec:
    #     tidx = np.searchsorted(self.magfield['time'], t)
    #     mf = self.magfield['values'][tidx, :]
    # dmats = compute_aniso_dmats(self.grid, smoothing=1, magflux=mf)

    if mask is None:
        mask = np.ones(grid.size, dtype=np.bool)
    dr = grid.dr
    dz = grid.dz
    npix = grid.nodes_num
    bmat = None
    atan2 = prepare_mag_data(magflux)
    bpar1, bper1, foo, bar = generate_anizo_matrix(grid, atan2, 1)
    bpar2, bper2, foo, bar = generate_anizo_matrix(grid, atan2, 2)
    
    [br, bz] = generate_iso_dmat(grid, derivative)
    bmat = [bpar1, bper1, bpar2, bper2]
    # correction of boundary conditions => zero derivative on borders !!
    # + ensure positive definiteness
    for i in range(4):
        bmat[i] = bmat[i] - spdiags(bmat[i].sum(axis=1).T, 0, npix, npix)
        ind = np.array(abs(bmat[i]).sum(axis=1).T == 0)
        corr = spdiags(ind, 0, npix, npix)*(br*dr+bz*dz) * spdiags(ind, 0, npix, npix)
        bmat[i] = bmat[i] + corr  # correction adds standard derivatives, probably not necessary
    bmat = [(bmat[0], bmat[1]), (bmat[2], bmat[3])]
    dmats_red = reduce_dmats(bmat, mask.flatten())
    return dmats_red


def prepare_mag_data(flux):
    """
    Calculates gradient map of magnetic flux function Psi(R,z) and returns
    arcus tangens of dPsi_y/-dPsi_x that is a suitable funciton for anizotropic
    diffusion matrix calculation
    
    Parameters
    ----------
    flux : numpy.ndarray
        array of Psi(R,z) evolution, with axes (z, R, t)
    
    Returns
    ------
    numpy.ndarray
        arcus tangens of Psi gradient
    """
    vgrad = np.gradient(flux[:, :])
    atan2 = np.arctan2(vgrad[1], -vgrad[0])
    return atan2


def px_norm(direc):
    """
    Computed (normalised?) distance between centers of nodes.
    """
    dn = np.sqrt(2)
    mod = np.asarray([s % 2 for s in direc])
    norms = np.zeros(np.shape(direc))
    norms[mod == 0] = 1
    norms[mod == 1] = dn
    return norms


def generate_anizo_matrix(grid, atan2, derivative):
    # TODO finish docstrings
    """
    Write prepared directions atan2 into the derivative matrix.
    Main magic of this algorithm. Rewrite arctan into the directions
    and decompose directions to parallel and oblique direction.

    Parameters
    ----------
    grid : Pixgrid
        Object with pixel grid gridinates
    atan2 : numpy.ndarray
        3D array of arcus tangents of Psi(R,z) evolution
    derivative : int
        derivative type identificator

    Returns
    -------
    bper : scipy.sparse
        sparse matrix with perpendicular derivatives
    bpar : scipy.sparse
        sparse matrix with parallel derivatives
    bpar_tmp: numpy.array
        (optional) dense matrix with parallel derivatives
    bper_tmp: numpy.array
        (optional) dense matrix with perpendicular derivatives
    """
    # obtain pixel dimensions
    nx = grid.nr
    ny = grid.nr
    npix = grid.nodes_num

    atan2 = atan2.flatten()
    
    # initiate 9point derivative matrices
    bper_tmp = np.zeros((npix, 9))
    bpar_tmp = np.zeros((npix, 9))
    center_tmp = 4
    
    # check for nans, potentially useful if vessel/separatrix truncated is supplied
    ind = ~np.isnan(atan2)
    atan2 = atan2[ind]
    n_ind = len(atan2)
    
    # decomposition of flux contour direction to two neighboring pixels
    for k in [0, 1]:
        # pixel with maximum contribution and px with second maximum contribution (45deg neighbour)
        direction = np.int_(np.mod(np.floor(atan2 / (np.pi/4) + k), 8))

        # array of reference indices
        dir_ = np.array((-1, 2, 3, 4, 1, -2, -3, -4), dtype=np.int)  # F like
        # dir_ = np.array((-3, -2, 1, 4, 3, 2, -1, -4), dtype=np.int)  # C like
        # C  [-4, -3, -2]       F  [-4, -1,  2]
        #    [-1,  0,  1]          [-3,  0,  3]
        #    [ 2,  3,  4]          [-2,  1,  4]
        # dir_ = np.array((-1, ny-1, ny, ny+1, 1, -ny+1, -ny, -ny-1), dtype=np.int)  # is compressed form
        
        next_ = np.squeeze(dir_[direction])
        
        # saw function, MAIN MAGIC, first steps to find projections of
        # the direction to the two neighboring pixels
        
        arelativ = np.abs(np.pi/4 - np.mod(atan2 + np.pi/4, np.pi/2))

        # obligue direction, zoom => ugly hack (works :)
        k1 = np.sin(2*arelativ)
        # direction paralel with axes
        k2 = np.cos(2*arelativ)    

        a = np.zeros(n_ind)
        ind_mod2 = np.bool_(np.mod(direction, 2))
        a[ind_mod2] = k1[ind_mod2]
        a[~ind_mod2] = k2[~ind_mod2]
        forind = center_tmp + next_
        backind = center_tmp - next_
        
        # Assign a value to specific pixels depending on the desired difference scheme:
        # 1 and 2 are just depending on the axis direction, 3 is second derivative and
        # 4 is central derivative
        # normalise pixels
        
        if derivative == 1:                        
            bper_tmp[ind, forind] = a/px_norm(forind)
        elif derivative == 2:
            bper_tmp[ind, backind] = a/px_norm(backind)
        elif derivative == 3:
            bper_tmp[ind, forind] = a/px_norm(forind)
            bper_tmp[ind, backind] = a/px_norm(backind)
        elif derivative == 4:
            bper_tmp[ind, forind] = a/px_norm(forind)
            bper_tmp[ind, backind] = -a/px_norm(backind)
        else:
            raise ValueError("Bad derivative type number, allowed {1,2,3,4}")

    # Constructing matrix with directions parallel to the magnetic field
    # from the matrix with directions perpendicular to magnetic field by essentially rotating
    # the outer pixels by 90deg
    ind = np.arange(8)
    pt1 = dir_[np.mod(ind, 8)]
    pt2 = dir_[np.mod(ind+2, 8)]
    # rotate directions by 90deg
    bpar_tmp[:, center_tmp + pt2[ind]] = bper_tmp[:, center_tmp + pt1[ind]]
    
    # normalisation and treatment for the central pixel: -1, 
    # except for central derivative where it is zero
    
    if derivative in (1, 2, 3):
        bper_tmp = sparse.spdiags(1 / (np.sum(bper_tmp, 1) + 0.000001), 0, npix, npix
                                  ) * bper_tmp
        bpar_tmp = sparse.spdiags(1 / (np.sum(bpar_tmp, 1) + 0.000001), 0, npix, npix
                                  ) * bpar_tmp
        bper_tmp[:, center_tmp] = -1
        bpar_tmp[:, center_tmp] = -1
    elif derivative == 4:
        bper_tmp = sparse.spdiags(1 / (np.sum(np.abs(bper_tmp), 1)), 0, npix, npix
                                  ) * bper_tmp
        bpar_tmp = sparse.spdiags(1 / (np.sum(np.abs(bpar_tmp), 1)), 0, npix, npix
                                  ) * bpar_tmp
    else:
        raise ValueError("Bad derivative type number, allowed {1,2,3,4}")
    # final conversion to npix x npix diagonal sparse matrices used in the calculation
    # bpar = sparse.spdiags(bpar_tmp.T, (ny+1, ny, ny-1, 1, 0, -1, -ny+1, -ny, -ny-1), npix, npix).T
    # bper = sparse.spdiags(bper_tmp.T, (ny+1, ny, ny-1, 1, 0, -1, -ny+1, -ny, -ny-1), npix, npix).T
    bpar = sparse.spdiags(bpar_tmp.T, (nx + 1, 1, -nx + 1, nx, 0, -nx, nx - 1, -1, -nx - 1), npix, npix).T
    bper = sparse.spdiags(bper_tmp.T, (nx + 1, 1, -nx + 1, nx, 0, -nx, nx - 1, -1, -nx - 1), npix, npix).T
    bpar = sparse.csc_matrix(bpar)
    bper = sparse.csc_matrix(bper)
    return bpar, bper, bpar_tmp, bper_tmp


# TODO default value for derivative, change type from int to str
def generate_iso_dmat(grid, derivative=4):
    """
    Computes isotropic derivative matrix.

    Parameters
    ----------

    Returns
    -------

    """
    npix = grid.nodes_num
    nx = grid.nr
    br, bz = None, None
    # diagonal
    diam = eye(npix, npix)
    diar = spdiags((np.ones((1, npix))), 1, npix, npix, format='csc')
    # to reference right nearest neighbors
    dial = spdiags((np.ones((1, npix))), -1, npix, npix, format='csc')
    # reference left nearest neighbors
    diao = spdiags((np.ones((1, npix))), nx, npix, npix, format='csc')
    # reference upper nearest neighbors
    diau = spdiags((np.ones((1, npix))), -nx, npix, npix, format='csc')
    
    # selection of suitable derivative
    if derivative == 1:
        br = -diam + dial
        bz = -diam + diao
    elif derivative == 2:
        br = diam - diar
        bz = diam - diau
    elif derivative == 3:
        br = dial - 2*diam + diar
        bz = diao - 2*diam + diau
    elif derivative == 4:
        br = dial - diar
        bz = diao - diau
    br = br / grid.dr
    bz = bz / grid.dz
    return br, bz


def reduce_dmats(dmats, idx):
    """
    Creates reduced derivative matrices by cutting out rows and columns representing unwanted nodes.

    Parameters
    ----------
    dmats : tuples of scipy.sparse.csr_matrix
        tuple containing full derivative matrices
    idx : numpy.ndarray of bool
        array or mask array to select desired nodes

    Returns
    -------
    list of scipy.sparse.csr_matrix pairs
        derivative matrices with removed rows and columns representing nodes outside the bdr polygon (vacuum vessel)
    """
    dmats_red = []
    for pair in dmats:
        s1 = pair[0][idx, :][:, idx]
        s2 = pair[1][idx, :][:, idx]
        pair_red = (s1, s2)
        dmats_red.append(pair_red)
    return dmats_red



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains analytical geometry matrix generators. Requires shapely library.

Shapely module is imported after the function is called so that it is not a hard dependence.
These are significantly slower than numerical ones, but can use polygons instead of line of sights.
"""
import time
import warnings

import numpy as np


def shapely_line_mat(xchord, ychord, grid):
    """
    Generates geometry matrix using shapely module
    
    Parameters
    ----------
    xchord, ychord : numpy.ndarray
        chord coordinates with shape (#chords, 2)
    grid : RegularGrid
        class of tomotok module describing reconstruction area

    Returns
    -------
    gmat : dict
        holds geometry matrix and grid coordinates
    """
    warnings.warn('Shapely based gmat generators will be removed in the future.', FutureWarning)
    import shapely.geometry as sg
    nx, ny = grid.nr, grid.nz
    xg, yg = grid.r_border, grid.z_border
    st = time.time()
    nch = xchord.shape[0]
    # cdct = dict(chnl=np.arange(nch), y=grid.z_center, x=grid.r_center)
    gmat = dict()
    gmat['values'] = np.zeros((nch, ny, nx))
    gmat['z'] = grid.z_center
    gmat['r'] = grid.r_center
    gmat['type'] = 'line'
    lines = []
    complex_chord = xchord.shape[1] > 2
    for i in range(nch):
        if complex_chord:
            crop = (xchord[i] < xg[-1]) * (xchord[i] > xg[0])
        else:
            crop = np.ones(xchord.shape[1], dtype=np.bool)
        coords = np.stack((xchord[i, crop], ychord[i, crop]), axis=1)
        line = sg.LineString(coords)
        lines.append(line)
    for row in range(ny):
        for col in range(nx):
            pix = sg.box(xg[col], yg[row], xg[col+1], yg[row+1])
            for i, line in enumerate(lines):
                gmat['values'][i, row, col] = pix.intersection(line).length
        if row % 10 == 9:
            ela = time.time() - st
            print('finished row {}/{}'.format(row + 1, ny),
                  'elapsed time {:.2f}s'.format(ela))
    ela = time.time() - st
    print('Gmat generation time {:.2f}s'.format(ela),
          'average time per chord {:.0f}ms'.format(ela/nch*1000))
    return gmat

# TODO remove
def shapely_poly_mat(xchord, ychord, grid, ph=None, w=None):
    """
    Generates geometry matrix using polygon object from shapely module.
    
    Parameters
    ----------
    xchord, ychord : numpy.ndarray
        chord coordinates with shape (#chords, #2) or (#chords, #curve_points)
    grid : RegularGrid
        class of tomotok module describing reconstruction area
    ph : numpy.ndarray, optional
        matrix with pinholes coords with shape (#chords, 2)
        If no values are supplied, start points of chords are used.
    w : numpy.ndarray, optional
        contains widening coefficients for chords with shape (#chords,).
        If no values are supplied, locally predefined value is used.

    Returns
    -------
    dict
        geometry matrix and parameters
    """
    warnings.warn('Shapely based gmat generators will be removed in the future.', FutureWarning)
    import shapely.geometry as sg
    # TODO different widths for each detector
    print('Computing shapely polygon geometry matrix')
    xg = grid.r_border
    yg = grid.z_border
    nx = grid.nr
    ny = grid.nz
    st = time.time()
    nch = xchord.shape[0]
    complex_chord = xchord.shape[1] > 2
    if ph is None or not complex_chord:
        ph = np.stack([xchord[:, 0], ychord[:, 0]], axis=1)
    if w is None:
        const = 0.022 * np.ones(nch)
    elif nch != len(w):
        warnings.warn('Size of provided widening coefficients'
                      ' is different from number of channels.',
                      RuntimeWarning)
        const = np.ones(nch)
        const[:len(w)] = w[:len(const)]
    else:
        const = w
    gmat = dict()
    gmat['values'] = np.zeros((nch, ny, nx))
    gmat['z'] = grid.z_center
    gmat['r'] = grid.r_center
    gmat['type'] = 'poly'
    center = np.array([xg[-1] + xg[0], yg[-1] + yg[0]])/2
    lines = []
    strings = []
    for i in range(nch):
        # xcrop = (xchord[i] < np.max(xg)) * (xchord[i] > np.min(xg))
        # ycrop = (ychord[i] < np.max(yg)) * (ychord[i] > np.min(yg))
        # crop = xcrop * ycrop
        if complex_chord:
            crop = (xchord[i] < xg[-1]) * (xchord[i] > xg[0])
        else:
            crop = np.ones(xchord.shape[1], dtype=np.bool)
        chord = np.stack((xchord[i, crop], ychord[i, crop]), axis=1)
        string = sg.LineString(chord)
        strings.append(string)
        u = chord - ph[i]
        mask = np.dot(u, center - ph[i]) > 0
        u = u[mask]
        n = const[i] * np.fliplr(u) * np.array([-1, 1])
        c1 = chord[mask] + n
        c2 = np.flipud(chord[mask] - n)
        ph_dst = np.linalg.norm(c1 - ph[i], axis=1)
        if ph_dst[0] < ph_dst[-1]:
            coords = np.vstack([ph[i], c1, c2])
        else:
            coords = np.vstack([c1, ph[i], c2])
        poly = sg.Polygon(coords)
        lines.append(poly)
        if not poly.is_valid:
            warnings.warn('invalid polygon for line {}'.format(i),
                          RuntimeWarning)
        # try:
        #     assert poly.is_valid
        #     lines.append(poly)
        # except AssertionError:
        #     wm = ('invalid polygon for line {}'.format(i) +
        #           'Replacing polygon with linestring')
        #     string = sg.LineString(chord)
        #     lines.append(string)
        #     warnings.warn(wm, RuntimeWarning)
    # TODO gamt to gmat['values']
    for row in range(ny):
        for col in range(nx):
            pix = sg.box(xg[col], yg[row], xg[col+1], yg[row+1])
            for i, line in enumerate(lines):
                # isec = pix.intersection(line)
                # area = isec.area
                gmat['values'][i, row, col] = pix.intersection(line).area
        if row % 10 == 9:
            ela = time.time() - st
            print('finished row {}/{}'.format(row + 1, ny),
                  'elapsed time {:.2f}s'.format(ela))
    lengths = lines_len(strings, grid)
    for i in range(nch):
        gmat['values'][i] = gmat['values'][i] / distance_mat(grid, ph[i])
        norm = lengths[i] / gmat['values'][i].sum()
        gmat['values'][i] *= norm
    ela = time.time() - st
    print('Gmat generation time {:.2f}s'.format(ela),
          'average time per chord {:.0f}ms'.format(ela/nch*1000))
    return gmat


def distance_mat(grid, ph):
    """
    Computes distances of pixels centers from pinhole and returns them in a
    matrix form. Indexing starts with bottom row.
    
    Parameters
    ----------
    grid : RegularGrid
        class describing reconstruction area
    ph : numpy.ndarray
        contains pinhole coordinates
    
    Returns
    -------
    numpy.ndarray
        matrix with pixel center distances from pinhole
    """
    x, y = np.meshgrid(grid.r_center - ph[0], grid.z_center - ph[1])
    dst = np.sqrt(np.power(x, 2) + np.power(y, 2))
    return dst

def lines_len(lines, grid):
    """
    Computes lengths of chords inside whole pixel grid

    Parameters
    ----------
    lines : array of shapely.geometry.LineString
        Contains central lines from polygon chord
    grid : RegularGrid

    Returns
    -------
    numpy.ndarray
        array with lengths of chords in pixel grid
    """
    warnings.warn('Shapely based gmat generators and auxiliary functions will be removed in the future.', FutureWarning)
    import shapely.geometry as sg
    box = sg.box(grid.rmin, grid.zmin, grid.rmax, grid.zmax)
    lengths = np.zeros(len(lines))
    for i, line in enumerate(lines):
        lengths[i] = box.intersection(line).length
    return lengths


# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
import matplotlib.pyplot as plt
import numpy as np


def compute_power(recs, grid):
    """
    Computes total radiated power from whole reconstruction area.

    Parameters
    ----------
    grid : RegularGrid

    Returns
    -------
    numpy.ndarray
        Contains values of total radiated power for all time slices
    """
    nslices = recs.shape[0]
    pixvol = grid.nodevol.flatten()
    ptot = np.ones(nslices)
    for i in range(nslices):
        rec = recs.flatten()
        ptot[i] = rec.dot(pixvol)
    return ptot


def plot(recs, ts=0):
    """
    Moved from BaseMfr class
    """
    raise NotImplementedError('TBD')
    data = self.results['used_data']
    errors = self.results['errors']
    # chnls = data.chnl.values
    chnls = self.results['chnls']
    bcli = self.results['retrofit']
    recs = self.results['recs']

    asp = plt.figaspect(0.5)
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=asp)

    ax1.set_title('Recs')
    # ax1.plot(*self.diag.load_boundary_coord().T, color='w')
    im = ax1.imshow(recs[ts], origin='lower', extent=self.grid.extent)
    plt.colorbar(im, ax=ax1)

    ax2.set_title('Retrofit')
    ax2.bar(chnls, bcli[ts], color='C0', alpha=1, label=r'$\mathbf{T} \cdot \mathbf{g}$')
    ax2.errorbar(chnls, data[ts, :], errors[ts, :],
                    ls='', marker='+', capsize=3, c='k', alpha=1, label='data')
    ax2.legend()
    ax2.yaxis.tick_right()
    plt.show()
    return fig


# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Simple function used to generate gifs from a set of png images.

Requires imageio library. To be replaced by matplotlib's animation
"""
import os

import imageio


def gifmaker(shot, what, duration, loops, fldr='../tomography/Graphs/all/'):
    """

    Parameters
    ----------
    shot : int
        checks file name if it contains required shot number
    what : str
        a string required to be present in file name, e.g. 'rectangles'
    duration
    loops : int
    fldr : str
        url of folder with pngs

    Returns
    -------

    """
    startdir = os.getcwd()
    os.chdir(fldr)
    slices = []
    names = []

    gif = '{} {} loops {} animation.gif'.format(shot, what, loops)
    if os.path.exists(gif):
        os.remove(gif)
#    mp4 = '{} {} video.mp4'.format(shot, what)
#    if os.path.exists(mp4):
#        os.remove(mp4)
    
    for fname in os.listdir():
        if what in fname.lower() and str(shot) in fname:        
            names.append(fname)
    names.sort()
    for name in names:
        slices.append(imageio.imread(name))
    imageio.mimwrite(gif, slices, duration=duration, loop=loops)
#    imageio.mimwrite(mp4, slices)
    os.chdir(startdir)



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains class describing regularly spaced node grid
"""
from typing import Tuple

import numpy as np
from matplotlib.path import Path


class RegularGrid(object):
    """
    Rectangular grid of regularly spaced rectangles of same size

    Describes rectangular reconstruction grid consisting of regularly spaced toroidally symmetric nodes
    that have rectangular projection to reconstruction plane.

    Attributes
    ----------
    nr : int
        number of nodes along r (radial) axis
    nz : int
        number of nodes along z (vertical) axis
    nodes_num : int
        total number of nodes
    rmin, rmax : float
        limits of r axis
    zmin, zmax : float
        limits of z axis
    rlims, zlims : Tuple[float, float]
        grid limits along respective axis
    dr, dz : float
        dimension of node in r or z axis
    r_border, z_border : numpy.ndarray
        arrays containing r resp. z coordinates of node borders
    r_center, z_center : numpy.ndarray
        arrays containing r resp. z coordinates of node centers
    nodevol : numpy.ndarray
        Array containing volumes of nodes (voxel) with shape (nz, nr)
    """

    def __init__(self, nr: int, nz: int, rlims: Tuple[float, float], zlims: Tuple[float, float]) -> None:
        """
        Parameters
        ----------
        nr : int
            number of nodes along radial axis r
        nz : int
            number of nodes along vertical axis z
        rlims : tuple of float
            limits of r axis, (min, max)
        zlims : tuple of float
            limits of z axis, (min, max)
        """
        self.nr = nr
        self.nz = nz
        self.nodes_num = self.size
        self.rmin = rlims[0]
        self.rmax = rlims[1]
        self.zmin = zlims[0]
        self.zmax = zlims[1]
        self.dr = (self.rmax - self.rmin) / self.nr
        self.dz = (self.zmax - self.zmin) / self.nz
        self.r_border = np.linspace(self.rmin, self.rmax, self.nr + 1)
        self.z_border = np.linspace(self.zmin, self.zmax, self.nz + 1)
        self.r_center = self.r_border[:-1] + self.dr / 2
        self.z_center = self.z_border[:-1] + self.dz / 2
        return

    @property
    def nodevol(self) -> np.ndarray:
        """
        Volumes of individual nodes

        Returns
        -------
        numpy.ndarray
            matrix with volume values for each node with shape (nz, nr)
        """
        nodevol_r = 2 * np.pi * self.dr * self.dz * self.r_center
        nodevol = np.ones((self.nz, self.nr)) * nodevol_r
        return nodevol

    @property
    def centre(self) -> Tuple[float, float]:
        """Returns the mean of rmin, rmax and zmin, zmax"""
        return (self.rmax + self.rmin)/2, (self.zmax + self.zmin)/2

    @property
    def extent(self) -> Tuple[float, float, float, float]:
        """
        Grid extent in format for use in matplotlib

        Returns
        -------
        tuple of floats
            rmin, rmax, zmin, zmax
        """
        return self.rmin, self.rmax, self.zmin, self.zmax

    @property
    def size(self) -> int:
        return self.nr * self.nz

    @property
    def shape(self) -> Tuple[float, float]:
        return self.nz, self.nr

    @property
    def center_mesh(self) -> np.ndarray:
        return np.meshgrid(self.r_center, self.z_center)

    @property
    def rlims(self) -> Tuple[float, float]:
        return (self.rmin, self.rmax)

    @property
    def zlims(self) -> Tuple[float, float]:
        return (self.zmin, self.zmax)

    def __repr__(self):
        msg = 'Node grid with resolution {}h{}'.format(self.nr, self.nz)
        msg += ' and bounds ({};{})r, ({};{})z.'.format(*self.extent)
        return msg

    def is_inside(self, r: np.ndarray, z: np.ndarray) -> np.ndarray:
        """
        Determines whether node centers are inside given polygon.

        Parameters
        ----------
        r, z : numpy.ndarray
            Coordinate vectors of polygon

        Returns
        -------
        numpy.ndarray
            Mask matrix for pixgrid with True values for nodes inside polygon
        """
        rm, zm = np.meshgrid(self.r_center, self.z_center)
        rm, zm = rm.flatten(), zm.flatten()
        points = np.stack((rm, zm), axis=1)
        limiter_coords = np.stack((r, z), axis=1)
        p = Path(limiter_coords)
        grid_points = p.contains_points(points)
        inside = grid_points.reshape(self.shape)
        return inside

    def corners(self, mask: np.ndarray=None) -> np.ndarray:
        """
        Creates an array with r, z coordinates of node corners. 
        
        Corners are in clockwise order starting with top left corner.

        Parameters
        ----------
        mask : numpy.ndarray
            2D bool matrix for selecting nodes, shape (#z, #r)

        Returns
        -------
        numpy.ndarray
            corner coordinates for each node in reconstruction plane, shape (#z, #r, 4, 2)
        """
        corners = np.empty((*self.shape, 4, 2))
        # top left
        tl = np.meshgrid(self.r_border[:-1], self.z_border[1:])
        # top rigth
        tr = np.meshgrid(self.r_border[1:], self.z_border[1:])
        # bottom right
        br = np.meshgrid(self.r_border[1:], self.z_border[:-1])
        # bottom left
        bl = np.meshgrid(self.r_border[:-1], self.z_border[:-1])
        
        corners = np.stack((tl, tr, br, bl)).transpose(2, 3, 0, 1)
        if mask is not None:
            corners = corners[mask]
        return corners



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains classes and functions used for creating database interfaces.

 - base class Dsystem for a general diagnostic system
 - sublcass with magnetics handling typical for tokamaks 
 - simple hdf interface using h5py
"""
from .diag import Dsystem
from .support import to_hdf, from_hdf
from .tokamaks import Tokamak



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
WIP

Contains class that does data filtering and stuff. It can also handle error estimation.
"""
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage.filters import generic_filter1d
from scipy.interpolate import interp1d
from scipy.signal import medfilt, butter, lfilter
from scipy.stats import sem


class DataHandler(object):
    """
    Does filtering and stuff

    Attributes
    ----------
    chnls : np.ndarray
    time : np.ndarray
    data : np.ndarray
    errors : np.ndarray
    """

    def __init__(self, data, min_error=0):
        """

        Parameters
        ----------
        data : dict
            [description]
        min_error : float, optional
            [description], by default 0
        """
        self.chnls = data['chnl']
        self.time = data['time']
        self.data = data['values']
        self._original = data.copy()
        self.errors = np.zeros_like(self.data) + min_error
        return

    # @property
    # def errors(self):
    #     return self._errors

    # @errors.setter
    # def errors(self, err):
    #     try:
    #         assert err.shape == self.data.shape
    #         self.errors = err
    #     except AssertionError:
    #         raise ValueError('Errors must have same shape as data')

    def reset(self, min_error=0):
        self.data = self._original
        self.errors = np.zeros_like(self.data) + min_error

    def add_errors(self, err):
        """
        Parameters
        ----------
        err : 
        """
        self.errors += err
        # self.errors = np.sqrt(np.power(self.err, 2) + np.power(err, 2))

    def errors_scalar_multiply(self, scalar):
        try:
            assert isinstance(scalar, (int, float))
            self.errors *= scalar
        except AssertionError:
            raise ValueError('scalar must be instance of int or float')

    def median_filter(self, kernel_size=1, error_type='sem'):
        if kernel_size % 2 == 0:
            warn('Medfilt does not support even values of kernel size. Adding +1.')
            kernel_size += 1
        self._filtering_errors(error_type, kernel_size=kernel_size)
        self._broadcast_to_channels(medfilt, kernel_size=kernel_size)

    def average(self, avg_num, tvec):
        tvec = np.array(tvec, ndmin=1)
        data = self.data
        avg_data = np.zeros((tvec.size, self.chnls.size))
        stds = np.zeros_like(avg_data)
        for i in range(tvec.size):
            idx = np.searchsorted(self.time, tvec[i])
            start = max(0, idx - avg_num)
            dsel = data[idx:start:-1, :]
            avg_data[i, ...] = dsel.mean(axis=0)
            stds[i, ...] = dsel.std(axis=0)
        return avg_data, stds

    # def bpf(self, error_type='coef', freq_lim=None):
    #     kwargs = {
    #         'lowcut': self.params['filter'][1],
    #         'highcut': self.params['filter'][2],
    #         'fs': 1 / (np.mean(np.diff(tvec))),
    #         'order': 2}
    #
    #     data_cut = self.cut_data(data_raw)  # TODO
    #
    #     data = self.__iterate(data_cut, self.butter_bandpass_filter, kwargs)
    #
    #     errors = self.compute_data_errors(data, error_type)
    #
    #     data_out = self.interpolate(data)
    #
    #     return data_out, errors

    # def gauss(self, error_type='sem', data_smooth=0):
    #     kwargs = {
    #         'sigma': data_smooth,
    #         'mode': 'nearest',
    #     }
    #
    #     self._broadcast_to_channels(self.butter_bandpass_filter, kwargs)
    #
    #     errors = self.compute_data_errors(error_type)

    def _interpolate_data(self, tvec):
        """
        Interpolates data to given time vector

        Parameters
        ----------
        tvec : np.ndarray

        Returns
        -------
        ndarray
        """
        data_interp = np.zeros((tvec.size, self.chnls.size))
        for i in range(self.chnls.size):
            dataf = interp1d(self.time, self.data[:, i])
            data_interp[:, i] = dataf(tvec)
        return data_interp

    def _interpolate_errors(self, tvec):
        """
        Interpolates data to given time vector

        Parameters
        ----------
        tvec : np.ndarray

        Returns
        -------
        ndarray
        """
        err_interp = np.zeros((tvec.size, self.chnls.size))
        for i in range(self.chnls.size):
            errf = interp1d(self.time, self.errors[:, i])
            err_interp[:, i] = errf(tvec)
        return err_interp

    def interpolate(self, tvec):
        """
        Interpolates both data and errors to given time vector.

        Parameters
        ----------
        tvec : int, float, np.ndarray

        Returns
        -------
        data, err : ndarray
        """
        tvec = np.array(tvec, ndmin=1)
        if tvec.min() < self.time.min() or tvec.max() > self.time.max():
            raise ValueError('Interpolating outside data time interval.')
        data = self._interpolate_data(tvec)
        err = self._interpolate_errors(tvec)
        return data, err

    def _broadcast_to_channels(self, func, *args, **kwargs):
        """
        Modifies data by applying provided function to each channel

        Parameters
        ----------
        func : callable
        args
            positional arguments
        kwargs
            keywords arguments
        """
        for i in range(self.data.chnl.size):
            self.data[i, :] = func(self.data[i, :], *args, **kwargs)

    @staticmethod
    def _sem(iline, oline, width):
        """
        Static wrapper for generic_filter1d using scypy.stats.sem as filter function over width samples
        """
        for i in range(len(oline)):
            oline[i] = sem(iline[i:i + width])

    @staticmethod
    def _std(iline, oline, width):
        """
        Static wrapper for generic_filter1d using np.std as filter function over width samples
        """
        for i in range(len(oline)):
            oline[i] = np.std(iline[i:i + width])

    @staticmethod
    def _none(iline, oline):
        pass

    def _error_filter(self, func, kernel_size, **kwargs):
        errors = generic_filter1d(self.data, func, kernel_size, axis=1, **kwargs)
        self.add_errors(errors)

    # TODO reconsider explicit error computation
    # def error_std(self, kernel_size):
    #     errors = generic_filter1d(self.data, self._std, kernel_size, axis=1, extra_keywords={'width': kernel_size})
    #     # self.add_errors(errors)
    #     return errors
    #
    # def error_sem(self, kernel_size):
    #     errors = generic_filter1d(self.data, self._std, kernel_size, axis=1, extra_keywords={'width': kernel_size})
    #     # self.add_errors(errors)
    #     return errors

    def _filtering_errors(self, error_type='std', **kwargs):
        """
        Computes errors induced by filtering data.
        """
        size = 1
        err_kw = {}
        if error_type.lower() == 'std':
            try:
                size = kwargs['kernel_size']
            except KeyError:
                raise ValueError('Parameter `kernel_size` must be provided for std method')
            error_func = self._std
            err_kw = {'extra_keywords': {'width': size}}
        elif error_type.lower() == 'sem':
            try:
                size = kwargs['kernel_size']
            except KeyError:
                raise ValueError('Parameter `kernel_size` must be provided for sem method')
            err_kw = {'extra_keywords': {'width': size}}
            error_func = self._sem
        elif error_type.lower() == 'poisson':
            raise NotImplementedError()
        elif error_type.lower() == 'wo':
            error_func = self._none
        else:
            raise ValueError('Unsuported error_type')
        self._error_filter(error_func, kernel_size=size, **err_kw)

    def crop_data(self, tlim, extend_num=None):
        """
        Cuts data to requested interval

        Parameters
        ----------
        tlim : tuple
            contains limits of requested time axis
        extend_num : int, optional
            contains number of points outside tlim for filtering methods
            (default value is calculated to fit chosen filtering method) <- should be determined by user
        """
        if tlim[0] < self.time.min() or self.time.max() < tlim[1]:
            raise ValueError('Requested time range out of data time vector.')
        t0 = np.searchsorted(self.time, tlim[0])
        t1 = np.searchsorted(self.time, tlim[0], side='right')
        if extend_num is None:
            self.data = self.data[t0:t1, :]
            self.errors = self.errors[t0:t1, :]
        else:
            raise NotImplementedError('Extension of croping interval was not implemented.')
            # FIXME make the method work with data attribute, remove return statement
            # TODO reconsider extend_num
            # ind_t0 = np.where(self.tvec >= tlim[0])[0][0]
            # ind_t1 = np.where(self.tvec <= tlim[1])[0][-1]
            #
            # try:
            #     self.data = self.data[:, ind_t0 - extend_num:ind_t1 + extend_num + 1]
            # except IndexError:
            #     warn('Time range with extension is out of data time vector. Using maximal possible extension.')
            #     self.data = self.data[:, max(0, ind_t0 - extend_num):min(ind_t1 + extend_num + 1, len(data.t))]
            # if ind_t0 - extend_num < 0:
            #     n_add = extend_num - ind_t0
            #     data_start = np.matlib.repmat(data.values[:, 0], n_add, 1).T
            #     self.data.values = np.concatenate((data_start, data.values), axis=1)
            #     dt = data.t.values[ind_t0 + 1] - data.t.values[ind_t0]
            #     self.data.t.values = np.concatenate(
            #         (np.linspace(data.t.values[0] - n_add * dt, data.t.values[0] - dt, n_add), data.t.values))
            #
            # if ind_t1 + extend_num > len(data.t):
            #     n_add = extend_num + ind_t1 - data.shape[1] + 1
            #     data_end = np.matlib.repmat(data.values[:, -1], n_add, 1).T
            #     self.data.values = np.concatenate((data.values, data_end), axis=1)
            #     dt = data.t.values[ind_t1] - data.t.values[ind_t1 - 1]
            #     self.data.t.values = np.concatenate(
            #         (data.t.values, np.linspace(data.t.values[-1] + dt, data.t.values[-1] + dt * n_add), n_add))

    def downsample(self, n, average=False):
        """
        Keeps only each n-th time slice of data.

        Parameters
        ----------
        n : int
        average : bool, optional
            not implemented yet
        """
        if average:
            raise NotImplementedError()
        else:
            self.data = self.data[::n, :]
            self.errors = self.errors[::n, :]

    def butter_bandpass_filter(self, lowcut, highcut, fs, order=5):
        nyq = 0.5 * fs
        low = lowcut / nyq
        high = highcut / nyq
        b, a = butter(order, [low, high], btype='band')
        y = lfilter(b, a, self.data)
        return y

    def remove_nan_chnls(self):
        """
        Removes channels that contain at least one Nan value.
        """
        mask = np.isfinite(self.data).all(0)
        self.chnls = self.chnls[mask]
        self.data = self.data[:, mask]
        self.errors = self.errors[:, mask]

    def remove_chnls(self, chnls):
        """
        Removes specified channels from data

        Parameters
        ----------
        chnls : iterable of ints
        """
        # chnls = np.array(chnls, ndmin=1, dtype=np.int)
        chnls = np.unique(chnls)
        mask = np.isin(chnls, self.chnls)
        if not mask.all():
            warn('Channels not in data: {}'.format(chnls[~mask]))
        chnls = chnls[mask]
        idx = np.searchsorted(self.chnls, chnls)
        mask = np.ones_like(self.chnls, dtype=np.bool)
        mask[idx] = 0
        self.chnls = self.chnls[mask]
        self.data = self.data[:, mask]
        self.errors = self.errors[:, mask]

    def remove_zero_chnls(self):
        sm = self.data.sum(axis=0)  # along time axis
        mask = sm != 0
        self.chnls = self.chnls[mask]
        self.data = self.data[:, mask]
        self.errors = self.errors[:, mask]

    def remove_below(self, threshold):
        """
        Removes channels with average signal bellow given threshold

        Parameters
        ----------
        threshold : float
        """
        sm = self.data.sum(axis=0)
        mask = sm >= threshold * self.time.size
        self.chnls = self.chnls[mask]
        self.data = self.data[:, mask]
        self.errors = self.errors[:, mask]

    def fourier(self, freqs):
        """
        Transforms data to frequency domain using fourier transform

        Parameters
        ----------
        freqs : list
            values of frequencies used for new `time vector`

        Returns
        -------
        DataHandler
            new instance of DataHandler with transformed data and errors
        """
        raise NotImplementedError()

    def __getitem__(self, item):
        return self.data.__getitem__(item)

    def __repr__(self):
        msg = '<DataHandler>'
        msg += '\n' + self.chnls.__repr__() + self.time.__repr__()
        msg += '\n Data \n' + '\n'.join(self.data.__repr__())#.split('\n')[1:-3])
        msg += '\n Errors \n' + '\n'.join(self.errors.__repr__())#.split('\n')[1:-3])
        return msg

    # def sel(self, *args, **kwargs):
    #     # for i in ('chnl', 't'):
    #     #     try:
    #     #         indexers[i] = np.array(indexers[i], ndmin=1)
    #     #     except KeyError:
    #     #         pass
    #     self._data = self.data.sel(*args, **kwargs)
    #     self._errors = self.errors.sel(*args, **kwargs)
    #     # dh = DataHandler(data)
    #     # dh.errors = err
    #     # return dh

    # def isel(self, *args, **indexers):
    #     data = self.data.isel(*args, **indexers)
    #     err = self.errors.isel(*args, **indexers)
    #     dh = DataHandler(data)
    #     dh.errors = err
    #     return dh

    def plot_chnls(self, chnls=None):
        if chnls is None:
            chnls = self.chnls
        fig = plt.figure()
        for ch in chnls:
            data = self.data[:, ch]
            err = self.errors[:, ch]
            plt.fill_between(self.time, data - err, data + err, alpha=0.5, interpolate=True)
            plt.plot(self.time, data)
        return fig

    def plot_time(self, time=0):
        fig = plt.figure()
        tidx = np.searchsorted(self.time, time)
        data = self.data[tidx, :]
        err = self.errors[tidx, :]
        plt.errorbar(self.chnls, data, err, ls='', marker='+', capsize=3, c='k', alpha=1)
        plt.bar(self.chnls, data)
        return fig



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Inversion algorithms independent on the rest of the package

Currently implemented
 - Minimum Fisher Regularisation (MFR)
 - Linear Algebraic Methods (LAME)
 - Biorthogonal Basis Decomposition (BOB)
"""

from .bob import Bob, SparseBob, SimpleBob, CholmodBob
from .lame import SvdFastAlgebraic, GevFastAlgebraic
from .mfr import Mfr, CholmodMfr



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains frame widget subclasses prepared for use as selection menu in different GUI apps.
"""
import tkinter as tk
from tkinter import ttk


class AnalyzerMenu(ttk.Frame):
    """
    TTK Frame based class containing options for tomography.
    """

    def __init__(self, parent=None):
        ttk.Frame.__init__(self, parent)  # padding = (3,3,3,3))
        self.grid(sticky='WE')
        self.shot = ttk.Frame(self, padding=(3, 3, 12, 12), relief='solid')
        self.shot.grid(row=0, column=0, columnspan=2)
        self.tokvar = tk.StringVar()
        self.cboxlbl = ttk.Label(self.shot, text='Tokamak:')
        self.cboxlbl.grid(row=0, column=0, sticky='EW')
        toklist = ['JET', 'COMPASS']
        self.cbox = ttk.Combobox(self.shot, values=toklist, width=10,
                                 state='readonly',
                                 textvariable=self.tokvar)
        self.cbox.current(0)
        self.cbox['state'] = 'readonly'
        # self.cbox['state'] = 'disabled'
        self.cbox.grid(row=0, column=1, sticky='WE')
        shotrow = 0
        shotcol = 2
#        self.shotvar = tk.StringVar(value=92292)
#        self.shotvar = tk.StringVar(value=92400)
#        self.shotvar = tk.StringVar(value=91718)
        self.shotvar = tk.StringVar(value=94396)
        self.shotlbl = ttk.Label(self.shot, text='Shot:')
        self.shotent = ttk.Entry(self.shot, textvariable=self.shotvar,
                                 width=5)
        self.shotlbl.grid(row=shotrow, column=shotcol + 0, sticky='E')
        self.shotent.grid(row=shotrow, column=shotcol + 1, sticky='WE')
        self.reslbl = ttk.Label(self.shot, text='Resolution')
        self.reslbl.grid(row=1, column=0)
        self.rescbox = ttk.Combobox(self.shot, values=['55x100', '115x196', '110x200'], state='active')
        self.rescbox.grid(row=1, column=1)
        self.rescbox.set('110x200')
        self.formlbl = ttk.Label(self.shot, text='Format')
        self.formlbl.grid(row=1, column=2)
        self.formcbox = ttk.Combobox(self.shot, values=['.npz', 'hdf'], state='readonly')
        self.formcbox.grid(row=1, column=3)
        self.formcbox.set('hdf')
        self.tveclbl = ttk.Label(self.shot, text='Time interval')
        self.tveclbl.grid(row=2, column=1, sticky='W')
        tvecrow = 3
        self.stimevar = tk.StringVar(value='40')
#        self.stimevar = tk.StringVar(value='44.53')
        self.stimelbl = ttk.Label(self.shot, text='from:')
        self.stimeent = ttk.Entry(self.shot, textvariable=self.stimevar)
        self.etimevar = tk.StringVar(value='55')
#        self.etimevar = tk.StringVar(value='44.7')
        self.etimelbl = ttk.Label(self.shot, text='to:')
        self.etimeent = ttk.Entry(self.shot, textvariable=self.etimevar)
        self.stimelbl.grid(row=tvecrow, column=0, sticky='E')
        self.stimeent.grid(row=tvecrow, column=1, sticky='WE')
        self.etimelbl.grid(row=tvecrow, column=2, sticky='E')
        self.etimeent.grid(row=tvecrow, column=3, sticky='WE')
        self.typeframe = ttk.Frame(self, padding=(3, 3, 12, 12),
                                   relief='solid')
        self.typeframe.grid(row=1, column=0, sticky='NSWE')
        t = 0
        l = 0
        # TODO set default values for diagnostics here
        self.bolvar = tk.BooleanVar(value=1)
        self.ntrvar = tk.BooleanVar(value=0)
        self.sxrvar = tk.BooleanVar(value=0)
        self.t0 = ttk.Label(self.typeframe, text='Load:')

        self.chbbol = ttk.Checkbutton(self.typeframe, variable=self.bolvar,
                                      onvalue=True, text='Bolometers')

        self.chbntr = ttk.Checkbutton(self.typeframe, variable=self.ntrvar,
                                      onvalue=True, text='Neutrons')

        self.chbsxr = ttk.Checkbutton(self.typeframe, variable=self.sxrvar,
                                      onvalue=True, text='SXR')
        self.t0.grid(row=t, column=l, sticky=(tk.E, tk.W))
        self.chbbol.grid(row=t + 1, column=l + 0, sticky=(tk.E, tk.W))
        self.chbntr.grid(row=t + 2, column=l + 0, sticky=(tk.E, tk.W))
        self.chbsxr.grid(row=t + 3, column=l + 0, sticky=(tk.E, tk.W))
        #        for i in range(3):
        #            self.columnconfigure(i, weight = 1)

        # data processing frame
        dprocrow = 1
        dproccol = 1
        dprocspan = 1
        self.dprocframe = ttk.Frame(self, padding=(3, 3, 12, 12),
                                    relief='solid')
        self.dprocframe.grid(row=dprocrow, column=dproccol,
                             sticky='NSWE',
                             columnspan=dprocspan)
        self.dprocframe.columnconfigure(0, weight=1)
        self.dprocframe.columnconfigure(1, weight=1)
        self.dproclbl = ttk.Label(self.dprocframe, text='Data processing:')
        self.dproclbl.grid(row=0, column=0, sticky='W')
        self.rgbvar = tk.BooleanVar(value=0)
        self.pwrvar = tk.BooleanVar(value=1)
        self.recvar = tk.BooleanVar(value=0)
        self.prfvar = tk.BooleanVar(value=0)
        self.cntvar = tk.BooleanVar(value=0)
        self.rgbchb = ttk.Checkbutton(self.dprocframe, text='CMY plot',
                                      variable=self.rgbvar, onvalue=True)
        self.rgbchb.grid(row=1, column=0, sticky='W')
        self.pwrchb = ttk.Checkbutton(self.dprocframe, text='Power plot',
                                      variable=self.pwrvar, onvalue=True)
        self.pwrchb.grid(row=2, column=0, sticky='W')
        self.recchb = ttk.Checkbutton(self.dprocframe, text='Rectangles',
                                      variable=self.recvar, onvalue=True)
        self.recchb.grid(row=3, column=0, sticky='W')
        self.prfchb = ttk.Checkbutton(self.dprocframe, text='Profiler',
                                      variable=self.prfvar, onvalue=True)
        self.prfchb.grid(row=4, column=0, sticky='W')
        self.cntchb = ttk.Checkbutton(self.dprocframe, text='Contours',
                                      variable=self.cntvar, onvalue=True)
        self.cntchb.grid(row=5, column=0, sticky='W')


class ExpandMenu(ttk.Frame):
    """
    Executed after main 'Do' method.
    Makes buttons for recreating closed graph windows.
    """

    def __init__(self, parent, **kw):
        ttk.Frame.__init__(self, **kw)

        self.explbl = ttk.Label(self, text='Create window:')
        self.explbl.grid(row=0, column=0, columnspan=2, sticky='W')
        #        for i in range(6):
        #            self.columnconfigure(i, weight = 1)
        bwid = 7
        self.bolbtn = ttk.Button(self, text='bol', width=bwid,
                                 command=parent.make_bol)
        self.bolbtn.grid(row=1, column=0)
        if not parent.menu.bolvar.get():
            self.bolbtn.state(['disabled'])
        self.ntrbtn = ttk.Button(self, text='ntr', width=bwid,
                                 command=parent.make_ntr)
        self.ntrbtn.grid(row=1, column=1)
        if not parent.menu.ntrvar.get():
            self.ntrbtn.state(['disabled'])
        self.sxrbtn = ttk.Button(self, text='sxr', width=bwid,
                                 command=parent.make_sxr)
        self.sxrbtn.grid(row=1, column=2)
        if not parent.menu.sxrvar.get():
            self.sxrbtn.state(['disabled'])
        self.mrfbtn = ttk.Button(self, text='mrf', width=bwid,
                                 command=parent.make_mrf)
        self.mrfbtn.grid(row=1, column=4)
        self.rgbbtn = ttk.Button(self, text='rgb', width=bwid,
                                 command=parent.make_rgb)
        self.rgbbtn.grid(row=2, column=0)
        self.pwrbtn = ttk.Button(self, text='pwr', width=bwid,
                                 command=parent.make_pwr)
        self.pwrbtn.grid(row=2, column=1)
        self.recbtn = ttk.Button(self, text='rec', width=bwid,
                                 command=parent.make_rec)
        self.recbtn.grid(row=2, column=2)
        self.prfbtn = ttk.Button(self, text='prf', width=bwid,
                                 command=parent.make_prf)
        self.prfbtn.grid(row=2, column=3)
        self.cntbtn = ttk.Button(self, text='cnt', width=bwid,
                                 command=parent.make_cnt)
        self.cntbtn.grid(row=2, column=4)
        self.masterbtn = ttk.Button(self, text='master',
                                    command=parent.make_mc, width=bwid + 3)
        self.masterbtn.grid(row=3, column=0, columnspan=2, sticky='W')


class SelectorMenu(ttk.Frame):
    def __init__(self, parent=None):
        ttk.Frame.__init__(self, parent)  # padding = (3,3,3,3))
        self.shot = ttk.Frame(self, padding=(3, 3, 12, 12), relief='solid')
        self.shot.grid(row=0, column=0, columnspan=2)
        tvecrow = 0
        self.tveclbl = ttk.Label(self.shot, text='Time interval')
        self.tveclbl.grid(row=tvecrow, column=1, sticky='W')
        self.stimevar = tk.StringVar(value='0')
        self.stimelbl = ttk.Label(self.shot, text='from:')
        self.stimeent = ttk.Entry(self.shot, textvariable=self.stimevar)
        self.etimevar = tk.StringVar(value='2')
        self.etimelbl = ttk.Label(self.shot, text='to:')
        self.etimeent = ttk.Entry(self.shot, textvariable=self.etimevar)
        self.stimelbl.grid(row=tvecrow+1, column=0, sticky='E')
        self.stimeent.grid(row=tvecrow+1, column=1, sticky='WE')
        self.etimelbl.grid(row=tvecrow+1, column=2, sticky='E')
        self.etimeent.grid(row=tvecrow+1, column=3, sticky='WE')
        self.typeframe = ttk.Frame(self, padding=(3, 3, 12, 12),
                                   relief='solid')
        self.typeframe.grid(row=1, column=0, sticky='NSWE')
        t = 0
        l = 0
        self.var1 = tk.StringVar()
        self.var2 = tk.StringVar()
        self.var3 = tk.StringVar()
        self.varm = tk.StringVar()

        self.t0 = ttk.Label(self.typeframe, text='Load:')
        self.mflbl = ttk.Label(self.typeframe, text='Magnetics:')

        self.ent1 = ttk.Entry(self.typeframe, textvariable=self.var1)
        self.ent2 = ttk.Entry(self.typeframe, textvariable=self.var2)
        self.ent3 = ttk.Entry(self.typeframe, textvariable=self.var3)
        self.entm = ttk.Entry(self.typeframe, textvariable=self.varm)

        self.txt1 = ttk.Button(self.typeframe, text='sel', command=parent.select1, width=3)
        self.txt2 = ttk.Button(self.typeframe, text='sel', command=parent.select2, width=3)
        self.txt3 = ttk.Button(self.typeframe, text='sel', command=parent.select3, width=3)
        self.txtm = ttk.Button(self.typeframe, text='sel', command=parent.selectm, width=3)

        self.t0.grid(row=t, column=l, sticky=(tk.E, tk.W))

        self.ent1.grid(row=t + 1, column=l + 0, sticky=(tk.E, tk.W))
        self.ent2.grid(row=t + 2, column=l + 0, sticky=(tk.E, tk.W))
        self.ent3.grid(row=t + 3, column=l + 0, sticky=(tk.E, tk.W))

        self.txt1.grid(row=t + 1, column=l + 1, sticky=(tk.E, tk.W))
        self.txt2.grid(row=t + 2, column=l + 1, sticky=(tk.E, tk.W))
        self.txt3.grid(row=t + 3, column=l + 1, sticky=(tk.E, tk.W))

        self.mflbl.grid(row=t + 4, column=l, sticky=(tk.E, tk.W))
        self.entm.grid(row=t + 5, column=l + 0, sticky=(tk.E, tk.W))
        self.txtm.grid(row=t + 5, column=l + 1, sticky=(tk.E, tk.W))
        #        for i in range(3):
        #            self.columnconfigure(i, weight = 1)

        # data processing frame
        dprocrow = 1
        dproccol = 1
        dprocspan = 1
        self.dprocframe = ttk.Frame(self, padding=(3, 3, 12, 12),
                                    relief='solid')
        self.dprocframe.grid(row=dprocrow, column=dproccol,
                             sticky='NSWE',
                             columnspan=dprocspan)
        self.dprocframe.columnconfigure(0, weight=1)
        self.dprocframe.columnconfigure(1, weight=1)
        self.dproclbl = ttk.Label(self.dprocframe, text='Data processing:')
        self.dproclbl.grid(row=0, column=0, sticky='W')
        self.rgbvar = tk.BooleanVar(value=0)
        self.pwrvar = tk.BooleanVar(value=1)
        self.recvar = tk.BooleanVar(value=0)
        self.prfvar = tk.BooleanVar(value=0)
        self.cntvar = tk.BooleanVar(value=0)
        self.rgbchb = ttk.Checkbutton(self.dprocframe, text='CMY plot',
                                      variable=self.rgbvar, onvalue=True)
        self.rgbchb.grid(row=1, column=0, sticky='W')
        self.pwrchb = ttk.Checkbutton(self.dprocframe, text='Power plot',
                                      variable=self.pwrvar, onvalue=True)
        self.pwrchb.grid(row=2, column=0, sticky='W')
        self.recchb = ttk.Checkbutton(self.dprocframe, text='Rectangles',
                                      variable=self.recvar, onvalue=True)
        self.recchb.grid(row=3, column=0, sticky='W')
        self.prfchb = ttk.Checkbutton(self.dprocframe, text='Profiler',
                                      variable=self.prfvar, onvalue=True)
        self.prfchb.grid(row=4, column=0, sticky='W')
        self.cntchb = ttk.Checkbutton(self.dprocframe, text='Contours',
                                      variable=self.cntvar, onvalue=True)
        self.cntchb.grid(row=5, column=0, sticky='W')
        return



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Template class for general tomographic diagnostic system.
"""
import warnings

import h5py
import numpy as np

from pathlib import Path


class Dsystem(object):
    """
    Contains basic information about diagnostic system

    Attributes
    ----------
    chnl_index : dict
        dict maping channel to detector
    geometry_path : pathlib.Path
        path to location with saved geometrical data
    name : str
        name of diagnostic system
    n_dets : int
        number of cameras with separate calibration files, (to be removed)
    """

    def __init__(self, shot, geometry_path='', **kw):
        super().__init__()
        self.name = self.__class__.__name__
        self.shot = shot
        self.geometry_path = Path(geometry_path)
        self.chnl_index = {}
        self.calb_0 = None
        self.n_dets = None
        self.keywords = kw
        return

    @property
    def data_name(self):
        """
        Template data file name
        """
        name = type(self).__name__
        return '{nm}_{shot}'.format(nm=name, shot=self.shot)

    def download_data(self, **kwargs):
        """
        Downloads data from network/database using `source` method or database.

        Returns
        -------
        dict
        """
        raise NotImplementedError("Download data not defined in subclass")

    def load_data(self, loc):
        """
        Loads hdf file containing data.

        If loc is path to a directory, default data file name from property data_name is assumed.

        Parameters
        ----------
        loc : str or Path
            Path to file with signal data. Can be full path or path to folder if default name is used.

        Returns
        -------
        dict
            keys values, chnl, time
        """
        loc = Path(loc)
        data = dict()
        if loc.is_dir():
            loc = Path(loc) / self.data_name
        with h5py.File(str(loc), 'r') as array:
            for key in array:
                data[key] = array[key][...]
        return data

    # TODO integrate into fetch and remove?
    @staticmethod
    def save_data(loc, data):
        """
        Saves downloaded detector data to data_path

        Parameters
        ----------
        loc : str or pathlib.Path
            saving path with file name
        data : dict
            contains signal values, channel numbers and time stamps
        """
        loc = Path(loc)
        try:
            loc.parent.mkdir(parents=True, exist_ok=True)
            with h5py.File(str(loc), 'w') as hdf:
                for key in ['values', 'chnl', 'time']:
                    hdf[key] = data[key]
        except OSError as e:
            warnings.warn('Detector data were not saved: {}'.format(e))
        return

    def fetch_data(self, data_path='', source=None, save=False, force_dl=False):
        """
        Loads data from data storage specified by data_path or downloads data
        from network storage.

        Parameters
        ----------
        data_path : str or Path, optional
            default is working directory
        source : :obj:`string`, optional
            specifies method for data download
        save : bool, optional
            switches saving to local storage
        force_dl : bool, optional
            forces download of data

        Returns
        -------
        dict
        """
        floc = Path(data_path)
        try:
            assert force_dl is False
            data = self.load_data(floc)
        except (OSError, AssertionError):
            data = self.download_data(source=source)
            if save:
                self.save_data(floc / self.data_name, data)
        return data

    def get_calb(self):
        """
        Loads preset calibrations for detectors, if not specified returns array of ones with length n_dets.
        """
        warnings.warn(
            'Method get_calb not specified in Dsystem subclass. ' +
            'Using default get_calb, which returns calb_0 or ones.',
            Warning)
        # TODO add method for loading detector calibrations
        if self.calb_0 is None:
            self.calb_0 = np.ones(self.n_dets)
        return self.calb_0

    def load_los(self, loc=None):
        """
        Loads line of sights geometry from hdf file.

        Parameters
        ----------
        loc : str or Path, optional
            location of los file, by default geometry path is used

        Returns
        -------
        dict
            with coordinates of los start and end points with shape (2, #chords, 3)
        """
        los = {}
        if loc is None:
            los_path = self.geometry_path / self.name
        else:
            los_path = Path(loc)
        with h5py.File(los_path, 'r') as fl:
            for key in fl.keys():
                st = fl[key]['start'][:]
                dir = fl[key]['dirs'][:]
                los[key] = np.stack((st, st+dir), axis=0)
        return los
    
    def get_chord_geometry(self):
        """
        Loads chord geometry from `geometry_path`.

        Expected names of detectors are detector_[num]_[x,y].txt.
        Data for x and y coordinates are in separate files.
        Each row represents one detector.
        Number of rows and columns should be same in all files containing coordinates

        Returns
        -------
        xchords, ychords : numpy.ndarray
            arrays with coordinates of chord points with shape (#chords, #points)
        """
        warnings.warn('Support for LoS saving to txt will be removed. Transfer your coordinates to hdf.', FutureWarning)
        geometry_path = self.geometry_path
        ychords = np.zeros((0, 0))
        xchords = np.zeros((0, 0))
        nl = 0
        i = 0
#        print(geometry_path / 'detector_{}_x.txt'.format(i))
        while (geometry_path / 'detector_{}_x.txt'.format(i)).exists():
            xchord = np.loadtxt(str(geometry_path / 'detector_{}_x.txt'.format(i)))
            if xchord.shape[1] == 3:
                xchord[:, 1] = np.average(xchord[:, :2], axis=1)
                xchord = xchord[:, 1:3]
            try:
                xchords = np.append(xchords, xchord, axis=0)
            except ValueError:
                xchords = xchords.reshape(0, xchord.shape[1])
                xchords = np.append(xchords, xchord, axis=0)
                ychords = ychords.reshape(0, xchord.shape[1])
            ychord = np.loadtxt(str(geometry_path / 'detector_{}_y.txt'.format(i)))
            if ychord.shape[1] == 3:
                ychord[:, 1] = np.average(ychord[:, :2], axis=1)
                ychord = ychord[:, 1:3]
            ychords = np.append(ychords, ychord, axis=0)
            self.chnl_index[i] = np.arange(xchord.shape[0]) + nl
            nl = nl + xchord.shape[0]
            i += 1
        if nl == 0:
            raise IOError('Could not import any geometric chords.' +
                          'Check geometry path and format of geometry data. ' +
                          str(geometry_path / 'detector_{}_x(y).txt'.format(i))
                          )
        return xchords, ychords

    def get_pinholes(self):
        """
        Loads pinhole coordinates for each chord from `pinholes.txt` located
        in geometry folder.

        Returns
        -------
        ph : numpy.ndarray
            matrix containing pinhole x,y coordinates with shape (#chords, 2)
        """
        geometry_path = self.geometry_path
        ph = np.loadtxt(str(geometry_path / 'pinholes.txt'))
        return ph

    def get_chord_widening(self):
        """
        Tries to load widening coefficients from file. Returns None if IOError

        Returns
        -------
        array-like or None
            Contains widening coefficients for each channel

        """
        geometry_path = self.geometry_path
        try:
            widths = np.loadtxt(str(geometry_path / 'widening.txt'))
        except IOError:
            widths = None
        return widths

    def load_boundary_coord(self):
        """
        Load border coordinates from geometry path or from tokamak module.
        Sets boundary_coord attribute to loaded coords.

        Returns
        -------
        numpy.ndarray
            contains (R, z) coordinates of vacuum vessel cross section points, should have shape (#points, 2)
        """
        warnings.warn('load_boundary_coord will be deprecated by load_border', FutureWarning)
        bdr_path = self.geometry_path / 'border.txt'
        boundary_coord = self.load_border(self, bdr_path)
        if np.all(boundary_coord[:, 0] != boundary_coord[:, -1]):
            boundary_coord = np.vstack([boundary_coord, boundary_coord[0, :][np.newaxis, :]])
        return boundary_coord


    def load_border(self, loc=None):
        """
        Load border coordinates from geometry path or from tokamak module.
        Sets boundary_coord attribute to loaded coords.

        Parameters
        ----------
        loc : str or Path

        Returns
        -------
        numpy.ndarray
            contains (R, z) coordinates of vacuum vessel cross section, shape (#points, 2)
        """
        if loc is None:
            bdr_path = self.geometry_path / 'border.txt'
        else:
            bdr_path = Path(loc)
        if not bdr_path.exists():
            print('border coord not found at {}.'.format(bdr_path))
            bdr_path = self.geometry_path.parent / 'border.txt'
            print('Tyring to load from {}'.format(bdr_path))
        bdr = np.loadtxt(str(bdr_path))
        return bdr

    # TODO move to Pixgrid?
    def compute_bd_mat(self, grid, loc=None):
        """
        Simple generation of boundary matrix 1 inside rec area, 0 outside

        Parameters
        ----------
        grid : RegularGrid

        Returns
        -------
        boundary_matrix : numpy.ndarray of bool
            matrix with true inside border, false otherwise, with shape (grid.nz, grid.nr)
        """
        limiter = self.load_border(loc)
        return grid.is_inside(limiter[:, 0], limiter[:, 1])



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
import matplotlib.pyplot as plt
import numpy as np

# TODO remove?
def projection_space(chords, centre=(0, 0), magnetics=None, maglvl=1):
    """

    Parameters
    ----------
    chords : tuple
    centre : tuple of float
    magnetics : dict
    maglvl : float, list of float

    Returns
    -------
    fig : matplotlib.figure
    """
    fig, ax = plt.subplots()
    xch, ych = chords
    dr = np.diff(xch)
    dz = np.diff(ych)
    nrm = np.sqrt(np.square(dr) + np.square(dz))
    p = np.zeros_like(xch)
    n = np.zeros_like(xch)
    p[:, 0] = centre[0] - xch[:, 0]
    p[:, 1] = centre[1] - ych[:, 0]
    n[:, 0] = -dz.flatten()
    n[:, 1] = dr.flatten()
    d = (n * p).sum(axis=1) / nrm.flatten()
    # d = np.abs(d)
    xi = np.arctan2(dz, dr).flatten()
    # d = np.where(xi > 0, d, -d)
    xi = np.where(xi > 0, xi, np.pi + xi)
    ax.plot(xi, d, ls='', marker='+', label='chords')
    if magnetics is not None:
        style = 'dotted'
        clr = 'k'
        vals = magnetics['values']
        x = magnetics['x']
        y = magnetics['y']
        cnt = plt.contour(x, y, vals, levels=maglvl)
        segs = cnt.allsegs
        for coll in cnt.collections:
            coll.remove()
        tcnt = []
        for seg in segs:
            if seg == []:
                tcnt.append(seg)
            else:
                # TODO fix for contours created by multiple segments of lines
                seg = seg[0]
                dr = seg[:, 0] - centre[0]
                dz = seg[:, 1] - centre[1]
                r = np.sqrt(np.square(dr) + np.square(dz))
                mxi = np.arctan2(dz, dr)
                rnum = -mxi.argmax()  # roll to avoid jump in xi angle
                mxi = np.roll(mxi, rnum)
                r = np.roll(r, rnum)
                uidx = np.where(mxi >= 0)[0]
                lidx = np.where(mxi < 0)[0]
                ax.plot(mxi[uidx], r[uidx], color=clr, ls=style)
                ax.plot(np.pi + mxi[lidx], -r[lidx], color=clr, ls=style)
                tcrd = np.stack((mxi, r), axis=1)
                tcnt.append(tcrd)
        ax.plot([], color=clr, ls=style, label='magnetics')

    ax.set_xlabel(r'Angle $ \xi $ [rad]')
    ax.set_ylabel('Length p [m]')
    ax.legend()
    # ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    # ax.set_xlim([0, 180])
    return fig



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Module with numeric generators of geometry matrix.
"""
import time
from warnings import warn

import numpy as np
import scipy.sparse as sparse
from matplotlib.path import Path

from .grids import RegularGrid


def sparse_line_3d(rchord, vchord, grid, ychord=None, step=1e-3, rmin=None):
    """
    Computes geometry matrix using simple numerical integration algorithm.

    Assumes toroidally symmetric reconstruction nodes. 
    Optimized version working with sparse matrices.

    Parameters
    ----------
    rchord : array-like
        radial coordinates of chord start and end points
    vchord : array-like
        vertical coordinates of chord start and end points
    grid : RegularGrid
    ychord : array_like, optional
        if specified, 3D Cartesian coordinates are assumed and rchord represents coordinates of x axis
    step : float, optional
        Integration step in meters.
    rmin : float, optional
        Stop iteration if r is lower. Useful when chords may go through central column.

    Returns
    -------
    sparse.csr_matrix
    """
    if ychord is None:
        ychord = np.zeros_like(rchord)
    rchord = np.array(rchord, ndmin=2)
    vchord = np.array(vchord, ndmin=2)
    ychord = np.array(ychord, ndmin=2)
    ela = time.time()
    nch = rchord.shape[0]
    gmat = sparse.lil_matrix((nch, grid.size))
    dr = np.diff(rchord, axis=1)
    dv = np.diff(vchord, axis=1)
    dy = np.diff(ychord, axis=1)
    dst = np.sqrt(dr * dr + dy * dy + dv * dv)
    for i in range(nch):
        steps = int(dst[i] / step)
        x = np.linspace(rchord[i, 0], rchord[i, 1], steps)
        y = np.linspace(ychord[i, 0], ychord[i, 1], steps)
        z = np.linspace(vchord[i, 0], vchord[i, 1], steps)
        r = np.sqrt(x**2 + y**2)
        if rmin is not None:
            hit = np.any(r < rmin)
            if hit:
                idx = r.argmin()
                r = r[:idx]
                z = z[:idx]
        hist = np.histogram2d(r, z, bins=[grid.r_border, grid.z_border])
        row = hist[0].T * dst[i] / steps
        srow = sparse.coo_matrix(row.flatten())
        gmat[i] = srow
    ela = time.time() - ela
    print('Gmat generation time {:.2f}s'.format(ela),
          'average time per chord {:.0f}ms'.format(ela / nch * 1000))
    return gmat.tocsr()


def calcam_sparse_line_3d(pupil, dirs, grid, steps=1e-3, rmin=None, elong=1.):
    """
    Computes geometry matrix from calcam input using sparse_line_3d algorithm.

    Assumes that pupil and dirs coordinates are (x, y, z) = (horizontal, horizontal, vertical)

    Parameters
    ----------
    pupil : np.ndarray
        (x, y, z) coordinates of pupil position
    dirs : np.ndarray
        (#rows, #columns, 3) line of sight direction vector coordinates
    grid : tomotok.core.geometry.RegularGrid
        reconstruction grid
    step : float, optional
        Integration step in meters.
    rmin : float, optional
        cut lines of sight if they intersect cylinder with radius of rmin, by default 0.3
    elong : float, optional
        multiplier for direction vectors elongation

    Returns
    -------
    sparse.csr_matrix
    """
    if elong != 1.:
        dirs = elong * dirs
    dirs = dirs.reshape(-1, 3)
    xchords = np.ones((dirs.shape[0], 2)) * pupil[0]
    xchords[:, 1] += dirs[:, 0]
    ychords = np.ones((dirs.shape[0], 2)) * pupil[1]
    ychords[:, 1] += dirs[:, 1]
    zchords = np.ones((dirs.shape[0], 2)) * pupil[2]
    zchords[:, 1] += dirs[:, 2]
    gmat = sparse_line_3d(xchords, zchords, grid, ychords, steps, rmin)
    return gmat



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains functions that can create emissivity phantoms.

Both isotropic one or based on the shape of flux surfaces.
"""
import numpy as np


def iso_psi(nx, ny, span=1.5):
    """
    Creates matrix of artificial isotropic psi profile with border values for
    each axis equal to span.

    Parameters
    ----------
    nx : int
        Number of pixels on x axis
    ny : int
        Number of pixels on y axis
    span : float, optional
        Value of result on the center of border

    Returns
    -------
    numpy.ndarray
        Matrix with generated profile
    """
    x = np.linspace(-span, span, nx)
    y = np.linspace(-span, span, ny)
    mx, my = np.meshgrid(x, y)
    res = np.sqrt(mx * mx + my * my)
    return res


def gauss(x, w=.1, lim=1, amp=1, cen=0.):
    r"""
    Creates anisotropic gaussian artificial emissivity by 1D transform of x

    .. math::
        f = amp \left( \mathrm{e}^{-(x-cen)^2 / w } - \mathrm{e}^{-(lim-cen)^2 / w)} \right)

    Can be used on np.ndarray. Lim should be greater than cen.

    Parameters
    ----------
    x : float, array, np.ndarray
        Contains values to be transformed, usually psi
    w : float, optional
        width of gaussian profile    
    lim : float, optional
        minimal value of x where transform gives zero
        if x > lim emissivity is set to zero    
    amp : float, optional
        amplitude of gaussian profile, 
    cen : float, optional
        center of gaussian profile
        allows hollow profile generation when mapped on psi

    Returns
    -------
    numpy.ndarray
        Transformed values of x with same dimensions
    """
    tx = (x - cen) * (x - cen)
    tlim = (lim - cen) * (lim - cen)
    res = amp * (np.exp(-tx / w) - np.exp(-tlim / w))
    res = np.where(res < 0, 0, res)
    return res


def gauss_iso(nx, ny, span=1.2, w=.1, lim=1, amp=1, cen=0):
    """
    Creates isotropic gaussian distribution.
    See references for iso_psi and gauss
    """
    x = iso_psi(nx, ny, span)
    res = gauss(x, w, lim, amp, cen)
    return res


def polar_phase(x, num=3, shift=0):
    """
    Applies sine phase in radial angle direction to given profile x.

    Parameters
    ----------
    x : numpy.ndarray
        Profile for application of polar phase.
    num : int, optional
        number of periods per one rotation
    shift : float, optional
        initial phase shift

    Returns
    -------
    numpy.ndarray
        Matrix with applied polar phase
    """
    # TODO pixels do not have to be sqaures
    rstep = 1
    cstep = 1
    im = np.argwhere(x == np.min(x))
    im = np.around(np.average(im, 0))
    r, c = np.shape(x)
    mr, mc = np.meshgrid(np.arange(c), np.arange(r))
    dr = (mr - im[1]) * rstep
    dc = (mc - im[0]) * cstep
    dst = np.sqrt(dr ** 2 + dc ** 2)
    dst[int(im[0]), int(im[1])] = -1
    angle = np.arcsin(dc / dst)
    angle = np.where(np.sign(dr) > 0, np.pi - angle, angle)
    # TODO consider using absolute value of sin
    multip = np.sin(num * angle + shift)
    res = x * multip
    return res


def islands(psi, w=.01, lim=1, amp=1, cen=0.4, num=3, shift=0):
    """
    Creates island like phantom from given psi profile. See references for gauss and polar_phase.
    """
    res = gauss(psi, w, lim, amp, cen)
    res = polar_phase(res, num, shift)
    return res



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains classes and functions for user checking of some tomography algorithm parts

Examples
--------
Checking anisotropic derivative matrix computed from DataArray with magnetic flux surfaces `magnetic_flux`

>>> from tomotok.core.io import Pixgrid, Tokamak
>>> time = 1.2
>>> shot = 19925
>>> coord = Pixgrid(50, 100, (0.5, 1), (-0.5, 0.5))
>>> magf = Tokamak.interpolate_mag_field(magnetic_flux, coord)
>>> checker = DerivMatChecker(coord, magf)
>>> checker(time)

# TODO
Plotting geometry matrix stored at local drive

>>> from tomotok.utils.gmat import load_function  # not implemented yet
>>> gmat = load_function('path/to/gmat/file')
>>> fig = check_gmat(gmat)
"""
import matplotlib.pyplot as plt
import numpy as np

from tomotok.core.derivative import prepare_mag_data, generate_anizo_matrix


class DerivMatChecker(object):
    """
    Computes anisotropic matrix for a time slice from provided magnetic flux and plots its components.
    Supports interactive selection of matrix element using mouse.

    Parameters
    ----------
    coord : Pixgrid
    magfield : dict

    Attributes
    ----------
    fig : matplotlib.figure
        contains color mesh and countour plot of magnetic surfaces
    fig2 :  matplotlib.figure
        contains two subplots with parallel and perpendicular parts of derivative matrix
    """
    def __init__(self, coord, magfield):
        self.nx, self.ny = coord.nx, coord.ny
        self.coord = coord
        self.fluxes = magfield
        self.fig, self.ax = plt.subplots()
        self.fig2, self.ax2 = plt.subplots(figsize=(8, 5), nrows=1, ncols=2)

        self.ax.set_title('Magnetic flux')
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)

        self.ax2[0].set_title('Check par_derivative')
        self.ax2[0].set_xlabel('x [px]')
        self.ax2[0].set_ylabel('y [px]')

        self.ax2[1].set_title('Check per_derivative')
        self.ax2[1].set_xlabel('x [px]')
        self.ax2[1].set_ylabel('y [px]')
        self.ax2[1].yaxis.tick_right()
        return

    def __call__(self, time, der_type=4):
        idx = np.searchsorted(self.fluxes['tvec'], time)
        flux = self.fluxes['values'][idx]
        self.t = time
        self.ax.images.clear()
        self.ax.collections.clear()
        self.ax.imshow(flux, origin='bottom',
                       # extent=self.coord.extent
                       )
        self.ax.contour(flux, np.sqrt(np.linspace(0, 1.0, 20)),
                        # extent=self.coord.extent
                        )

        atan2 = prepare_mag_data(flux[:, :])
        bper, bpar, bpar_dense, bper_dense = generate_anizo_matrix(self.coord, atan2, der_type)

        self.deriv1 = bper_dense.reshape(self.nx * self.ny, 3, 3)[:, :, :]
        self.deriv2 = bpar_dense.reshape(self.nx * self.ny, 3, 3)[:, :, :]
        # Central pixel index
        ind = int(self.coord.ny / 2 * self.coord.nx + self.coord.nx / 2)

        self.slices, rows, cols, = self.deriv1.shape
        sliced = self.deriv1[ind, :, :]
        self.im1 = self.ax2[0].imshow(sliced, cmap='Greys')

        self.slices, rows, cols, = self.deriv2.shape
        sliced = self.deriv2[ind, :, :]
        self.im2 = self.ax2[1].imshow(sliced, cmap='Greys')

        self.ax2[0].texts.clear()
        self.ax2[1].texts.clear()
        self.annotate1_list = {}
        self.annotate2_list = {}

        for i, j in np.ndindex(self.deriv1[ind, :, :].shape):
            # print(ind, i, j)
            annotate1_tmp = self.ax2[0].annotate('{:0.2}'.format(self.deriv1[ind, i, j]), (j, i), color='r')
            annotate2_tmp = self.ax2[1].annotate('{:0.2}'.format(self.deriv2[ind, i, j]), (j, i), color='r')
            self.annotate1_list[(i, j)] = annotate1_tmp
            self.annotate2_list[(i, j)] = annotate2_tmp

        plt.tight_layout()

        self.fig.canvas.draw()
        self.fig2.canvas.draw()
        return

    def update(self, x, y):
        """
        Updates plots for current pixel coordinate
        """
        self.ax2[0].set_title('Bpar, x={},y={}'.format(x, y))
        self.ax2[1].set_title('Bper, x={},y={}'.format(x, y))

        ind = int(y * self.coord.nx + x)

        sliced = self.deriv1[ind, :, :]
        self.im1.set_data(sliced)
        # self.im1.axes.figure.canvas.draw()

        sliced = self.deriv2[ind, :, :]
        self.im2.set_data(sliced)

        for i, j in np.ndindex(self.deriv1[ind, :, :].shape):
            self.annotate1_list[(i, j)].set_text('{:0.2}'.format(self.deriv1[ind, i, j]))
            self.annotate2_list[(i, j)].set_text('{:0.2}'.format(self.deriv2[ind, i, j]))
        self.fig2.canvas.draw()

    def onclick(self, event):
        """
        Gets coordinate from the click on the main flux plot
        """
        # print('Returned indices')
        # print(event.xdata, event.ydata)
        # print('mapping back:')
        x = int(np.round(event.xdata))
        y = int(np.round(event.ydata))
        # tx = "Y: {}, X: {}".format(y, x)
        # print(tx)
        self.update(x, y)


def check_gmat(gmat):
    """
    Checks geometry matrix by summing all channels and plotting.

    Parameters
    ----------
    gmat : dict

    Returns
    -------
    fig : matplotlib.pyplot.figure
    """
    raise NotImplementedError('Requires gmat class, grid descritpion in gmat attrs or grid as parameter')
    fig, ax = plt.subplots()
    sgmat = gmat.sum(0)
    s = ax.imshow(sgmat, origin='bottom')
    plt.colorbar(s)
    ax.set_ylabel('y grid [px]')
    ax.set_xlabel('x grid [px]')
    plt.show()
    return fig



import os.path as path
from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# parse the package version number
with open(path.join(path.dirname(__file__), 'tomotok/core/VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    name='tomotok',
    version=version,
    license='EUPL 1.2',
    author='Jakub Svoboda, Jordan Cavalier, Ondrej Ficker, Martin Imrisek et al.',
    # author_email='svoboda@ipp.cas.cz',
    description='Collection of algorithms used for tokamak plasma tomography',
    long_description=long_description,
    long_description_content_type='text/markdown',
    # url='',
    packages=find_packages(),
    namespace_packages=['tomotok'],
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=['numpy>=1.13.3', 'scipy>=1.1.0', 'matplotlib>=2.2.2', 'h5py>=2.7.1'],
)



# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Structure of classes is based on algorithms proposed by T. Odstrcil however without sparse optimization

T. Odstrcil et al., "Optimized tomography methods for plasma emissivity reconstruction at the
ASDEX Upgrade tokamak," Rev. Sci. Instrum., 87(12), 123505.
"""
from warnings import warn

import numpy as np
import scipy.sparse as sparse
from scipy.stats.mstats import mquantiles
from scipy.sparse.linalg import eigsh


class Algebraic(object):
    """
    A base class for algebraic inversion methods using linear regularisation.

    Attributes
    ----------
    u : numpy.ndarray
        decomposition matrix with shape (#channels, #channels)
    s : numpy.ndarray
        diagonal from a decomposition matrix S with shape (#channels, )
    v : numpy.ndarray
        decomposition matrix with shape (#nodes, #channels)    
    """
    def __init__(self):
        self.u = None
        self.s = None
        self.v = None
        self.alpha = None
        return

    def invert(self, data, gmat, regularisation, method, num=None):
        """
        Computes linear inversion using algebraic method.
        The inversion comprises of three stages:

            - decomposition (presolving) using only geometry and derivative matrices
            - searching for regularisation parameter
            - solving inversion using series expansion

        Parameters
        ----------
        data : numpy.ndarray
        gmat : numpy.ndarray
        regularisation : numpy.ndarray
            regularisation matrix
        method : str, optional
            method used for regularization parameter computation, see method find_alpha
        num : int, optional
            use only num most significant vectors in series expansion

        Returns
        -------
        numpy.ndarray
            reconstructed emissivity vector with shape (#pix,)
        """
        self.u, self.s, self.v, = self.decompose(gmat, regularisation)
        alpha = self.find_alpha(method)
        self.alpha = alpha
        g = self.series_expansion(alpha, data, num=num)
        return g

    # TODO pass regularisation matrix instead of derivatives
    def __call__(self, data, gmat, derivatives, errors, aniso=1, method=None, num=None):
        """
        Iterates all time slices in data and computes inversion using one of linear algebraic methods.

        Parameters
        ----------
        data : numpy.ndarray
            signal with shape (#time slices, #chnls)
        gmat : numpy.ndarray
            geometry matrix with shape (#chnls, #nodes)
        derivatives : list
            list of tuples with derivative matrix pairs
        errors : int, float or numpy.ndarray
            expected errors used to normalize data, with shape (#slices,), (#chnls,) or (#time slices, #chnls) 
        aniso : float, optional
            anisotropic factor making the first 
        method : str, optional
            specifies method used for regularization parameter computation, see method find_alpha
        num : int, optional
            sets number of largest vectors summed in the series expansion

        Returns
        -------
        numpy.ndarray
            results of inversion with shape (#timeslices, ny, nx)
        """
        nslices = data.shape[0]
        nchnls = data.shape[1]
        if nchnls != gmat.shape[0]:
            raise ValueError('Different number of channels in data and gmat')
        nnodes = gmat.shape[1]

        data_ndim = np.ndim(data)
        if data_ndim == 0:
            raise ValueError('Data must be at least an array of values.')
        elif data_ndim == 1:  # flat data, assume one time slice
            data = data.reshape(1, -1)
        elif data_ndim > 2:
            raise ValueError('Data array has too many dimension.')

        err_ndim = np.ndim(errors)
        if err_ndim == 0:  # constant errors
            errors = np.full_like(data, errors)
        elif err_ndim == 1:  # flat errors, check for matching shape
            if errors.size == nslices:
                errors = errors.reshape(-1, 1)
            elif errors.size == nchnls:
                errors = errors.reshape(1, -1)
            else:
                raise ValueError('Incompatible shape of errors {} with the data {}'.format(errors.shape, data.shape))
            errors = errors * np.ones_like(data)  # broadcast to right shape
        elif err_ndim > 2:
            raise ValueError('Errors array has too many dimensions.')

        if errors.shape != data.shape:
            raise ValueError('Data shape {} does not match errors shape {}.'.format(data.shape, errors.shape))
   
        data = data / errors

        res = np.empty((nslices, nnodes))

        reg = self.regularisation_matrix(derivatives, aniso)
        
        for i in range(data.shape[0]):
            signal = data[i, :].flatten()
            errs = errors[i, :]
            gmat_nrm = self.normalize_gmat(gmat, errs)
            #TODO sparse optimization?
            # error_sp = sparse.diags(1/errors[i, :])
            # gmat_nrm = error_sp.dot(gmat).toarray()
            res[i] = self.invert(signal, gmat_nrm, reg, method=method, num=num)
        return res
    
    # TODO make it an external function
    def regularisation_matrix(self, derivatives, aniso=1):
        """
        Computes regularisation matrix from derivatives matrices.

        Parameters
        ----------
        derivatives : _type_
            _description_
        aniso : int, optional
            _description_, by default 1

        Returns
        -------
        _type_
            _description_
        """
        # relative weighting
        w1 = aniso / (1 + aniso)
        w2 = 1 / (1 + aniso)
        n_der = len(derivatives)
        hs = np.zeros((n_der, *derivatives[0][0].shape))
        for i in range(n_der):
            # TODO
            tmp0 = derivatives[i][0].T.dot(derivatives[i][0])
            tmp1 = derivatives[i][1].T.dot(derivatives[i][1])
            hs[i, ...] = (w1 * tmp0 + w2 * tmp1).toarray()
        # h = np.sum(hs).toarray()
        h = hs.mean(axis=0)
        return h
    
    def normalize_gmat(self, gmat: np.ndarray, errors: np.ndarray) -> np.ndarray:
        """
        Normalizes gmat using estimated errors

        Parameters
        ----------
        gmat : np.ndarray
            geometry matrix
        errors : np.ndarray
            errors for given time slice shape (#channels,)

        Returns
        -------
        np.ndarray
            normalized geometry matrix
        """
        errors = np.diag(1/errors)
        gmat = errors.dot(gmat)
        return gmat

    def presolve(self, gmat, deriv):
        """
        .. deprecated :: 1.1
            Use :meth:`decompose`  
        """
        self.decompose(gmat, deriv)
        warn('Presolve method deprecated by decompose method since 1.1', DeprecationWarning)

    def decompose(self, gmat, regularisation):
        """
        Prepares matrices in standard form for series expansion

        Returns
        -------
        u, s, v : numpy.ndarray
        """
        raise NotImplementedError()

    def find_alpha(self, *args, **kwargs):
        """
        Finds regularisation parameter.

        Returns
        -------
        float
            regularisation parameter value
        """
        raise NotImplementedError()

    def series_expansion(self, alpha, signal, num=None):
        r"""
        Computes emissivity :math:`g` from decomposed vectors using

        .. math::
            \mathbf{g}(\alpha) = \sum_{i=1}^{m} \frac{k_{i} (\alpha)}{S_{ii}}
            \left( \mathbf{U}^T \cdot \mathbf{f} \cdot \tilde{\mathbf{V}} \right) {}_{*i},

        where :math:`k_i(\alpha)` are so called filtering factors computed using following formula

        .. math::
            k_{i}(\alpha) = \left(1 + \frac{\alpha}{S_{ii}^2} \right)^{-1}
        

        Parameters
        ----------
        alpha : float
            regularisation parameter
        signal : numpy.ndarray
            vector with channel signal
        num : int
            number of columns used for series expansion

        Returns
        -------
        numpy.ndarray
            results of inversion
        """
        s = self.s.reshape(1, -1)  # create row vector from diagonal matrix
        s_sq = np.square(s)
        filters = 1 / (1 + alpha / s_sq)
        tmp = filters / s * self.u.T.dot(signal) * self.v
        g = tmp[:, :num].sum(axis=1)
        return g


class FastAlgebraic(Algebraic):
    """
    A base class for linear algebraic methods using fast regularisation parameter estimation.
    
    The regularisation parameter estimate is based on the diagonal matrix obtained by decomposition.
    """
    def __init__(self):
        super().__init__()
        return

    def decompose(self, gmat, regularisation):
        raise NotImplementedError()

    def find_alpha(self, method: str='quantile'):
        """
        Finds regularisation parameter using linear estimate based on values of diagonal.

        Parameters
        ----------
        method : str {'mean', 'half', 'median', 'quantile', 'logmean}
            selects method for finding regularisation parameter value, default is quantile

        Returns
        -------
        float
            squared value found by estimation method
        """
        if method is None:
            method = 'quantile'
        if method == 'mean':
            alpha = self.s.mean()
        elif method == 'half':
            alpha = self.s.max() / 2
        elif method == 'median':
            alpha = mquantiles(self.s, prob=0.5, alphap=0, betap=1)[0]
        elif method == 'quantile':
            quant = 2 / np.e
            alpha = mquantiles(self.s, prob=quant, alphap=0, betap=1)[0]
        elif method == 'logmean':
            log_s = np.log10(self.s)
            alpha = np.power(10, log_s.mean())
        else:
            raise ValueError('Unrecognized option for regularisation parameter estimation: {}'.format(method))
        return alpha**2


class SvdFastAlgebraic(FastAlgebraic):
    def decompose(self, gmat, regularisation):
        l_mat = np.linalg.cholesky(regularisation)
        p = np.identity(l_mat.shape[0])
        l_inv = np.linalg.inv(l_mat)
        a = l_inv.dot(p).dot(gmat.T)
        u, s, vt = np.linalg.svd(a.T, full_matrices=False)
        v = vt.T
        v_tild = p.T.dot(l_inv.T).dot(v)
        return u, s, v_tild


class QrFastAlgebraic(FastAlgebraic):
    def __init__(self):
        raise NotImplementedError('Not yet finished')

    def decompose(self, gmat, regularisation):
        l_mat = np.linalg.cholesky(regularisation)
        p = np.identity(l_mat.shape[0])
        l_inv = np.linalg.inv(l_mat)
        a = l_inv.dot(p.dot(gmat.T))
        q1, d_roof, s = np.linalg.qr(a.dot(p))
        q2, r2 = np.linalg.qr(p.dot(s.T))
        m = d_roof.dot(r2.T).dot(np.linalg.inv(d_roof))
        r3, d3, q3 = np.linalg.qr(m)
        raise NotImplementedError('Not yet finished')


class GevFastAlgebraic(FastAlgebraic):
    def decompose(self, gmat, regularisation):
        """
        Decomposes geometry and regularisation matrices to form suitable for series expansion.
        
        Uses generalised eigenvalue decomposition scheme described by L. C. Ingesson in [GEV]_

        Parameters
        ----------
        gmat : numpy.ndarray
        regularisation : numpy.ndarray
            regularisation matrix with shape (#nodes, #nodes)

        Returns
        -------
        u, s, v

        References
        ----------
        .. [GEV] L.C. Ingesson, "The Mathematics of Some Tomography Algorithms Used at JET," JET Joint Undertaking, 2000
        """
        c = gmat.T.dot(gmat)
        c_csr = sparse.csr_matrix(c.T)
        reg_sparse = sparse.csc_matrix(regularisation)  # csc might be slightly faster that csr?
        s, ev = eigsh(c_csr, k=gmat.shape[0], M=reg_sparse)

        # flip to have eigenvalues and vectors sorted from largest to smallest
        s = s[::-1]
        ev = ev[..., ::-1]

        s_sqrt = np.sqrt(s)

        u = gmat.dot(ev) / s_sqrt
        v = s_sqrt * ev
        return u, s, v
